from configparser import ConfigParser

import tator

from hms_import.o2.upload import run_o2_upload
from hms_import.util import HmsLogHandler


def main(console_log_level: int, config_file: str):
    config = ConfigParser()
    config.read(config_file)
    host = config["Tator"]["Host"]
    token = config["Tator"]["Token"]
    tator_api = tator.get_api(host=host, token=token)
    with HmsLogHandler(tator_api=tator_api, console_log_level=console_log_level) as log_handler:
        log_handler.summary_image_id = run_o2_upload(tator_api=tator_api, config=config)



#!/usr/bin/env python3

import datetime
from glob import glob
import hashlib
import logging
import os
from pprint import pformat
from typing import Any, Dict, List, Tuple

import tator
from tator.openapi.tator_openapi import TatorApi
from tator.openapi.tator_openapi.models import CreateListResponse, UploadInfo
from tator.util._upload_file import _upload_file
from tqdm import tqdm


logger = logging.getLogger(__name__)


def _calculate_md5(file_path):
    hash_md5 = hashlib.md5()

    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)

    return hash_md5.hexdigest()


def get_start_end_times(path: str) -> Tuple[datetime.datetime, datetime.datetime]:
    directory = os.path.dirname(path)
    filename = os.path.basename(path)
    logger.info(f"Searching for toc_files in {directory=}")
    toc_files = list(glob(os.path.join(directory, f"*.video-toc")))
    logger.info(f"Searching for {filename=} in {toc_files=}")
    for toc_file in toc_files:
        with open(toc_file) as fp:
            for line in fp.readlines():
                decrypted_filename, start_time_ms, end_time_ms = line.strip().split(",")
                if decrypted_filename == filename:
                    return (
                        datetime.datetime.fromtimestamp(
                            int(start_time_ms) / 1000, tz=datetime.timezone.utc
                        ),
                        datetime.datetime.fromtimestamp(
                            int(end_time_ms) / 1000, tz=datetime.timezone.utc
                        ),
                    )
    raise RuntimeError(f"Could not get start and end times for {filename}!")


def _upload_video(
    tator_api: TatorApi,
    path: str,
    project_id: int,
    media_type: int,
    file_ids: List[int],
    section: str,
    vessel_name: str,
    skip_decrypt: bool,
) -> int:
    # Perform idempotency check
    filename = os.path.basename(path)
    paginator = tator.util.get_paginator(tator_api, "get_media_list")
    page_iter = paginator.paginate(project=project_id, name=filename, type=media_type)
    try:
        media = next(media for page in page_iter for media in page)
    except (RuntimeError, StopIteration):
        # Bug in paginator will raise RuntimeError if zero entities are returned
        pass
    else:
        logger.info(
            "Found existing media %d with same file name, skipping media creation%s",
            media.id,
            "" if skip_decrypt else " and upload",
        )
        return media.id

    attrs: Dict[str, Any] = {"Vessel Name": vessel_name}
    if skip_decrypt:
        toc_start, toc_end = get_start_end_times(path)
        attrs["decrypted_name"] = filename
        attrs["toc_start"] = toc_start
        attrs["toc_end"] = toc_end
    else:
        # Upload encrypted file to Tator
        logger.info("Uploading %s", filename)
        response = None
        try:
            for progress, response in _upload_file(
                tator_api, project_id, path=path, filename=filename
            ):
                logger.info("Upload progress: %.1f%%", progress)
        except Exception as exc:
            raise RuntimeError(f"Raised exception while uploading '{filename}', skipping") from exc

        if not isinstance(response, UploadInfo) or response.key is None:
            raise RuntimeError(f"Did not upload '{filename}', skipping")

        logger.info("Uploading %s successful!", filename)
        attrs["encrypted_path"] = response.key
        attrs["related_files"] = ",".join(str(file_id) for file_id in file_ids)

    # Create a media object containing the key to the uploaded file
    media_spec = {
        "type": media_type,
        "section": section,
        "name": filename,
        "md5": _calculate_md5(path),
        "attributes": attrs,
    }

    logger.info(
        "Creating media object in project %d with media_spec=%s", project_id, pformat(media_spec)
    )
    try:
        response = tator_api.create_media_list(project_id, [media_spec])
    except Exception as exc:
        raise RuntimeError(f"Could not create media with {project_id=} and {media_spec=}") from exc

    return response.id[0] if isinstance(response, CreateListResponse) and response.id else 0


def upload_videos(
    *,
    tator_api: TatorApi,
    directory: str,
    media_ext: str,
    project_id: int,
    media_type: int,
    section: str,
    file_ids: List[int],
    vessel_name: str,
    skip_decrypt: bool,
) -> List[int]:
    """
    Finds all encrypted video files in `directory`, uploads them to Tator, and creates a media
    object referencing the upload. A future algorithm will be responsible for decrypting and
    transcoding them. Disallows use of positional arguments.

    :tator_api: The TatorApi object to use for interactions with Tator
    :param token: The Tator API token to use for authentication.
    :type token: str
    :param directory: The directory to search for encrypted video files.
    :type directory: str
    :param media_ext: The extension of the encrypted video files.
    :type media_ext: str
    :param project_id: The integer id of the project to upload the videos to.
    :type project_id: int
    :param media_type: The integer id of the media type to create.
    :type media_type: int
    :param section: The section to upload the media to.
    :type section: str
    :param file_ids: The list of file ids to associate uploaded media with.
    :type file_ids: List[int]
    :param vessel_name: The dict of shared attributes for all media.
    :type vessel_name: Dict[str, str]
    :param skip_decrypt: If True, skips uploading encrypted files and assumes files in `directory`
                         are previously decrypted legacy files
    :type skip_decrypt: bool
    """
    if skip_decrypt:
        file_list = glob(os.path.join(directory, f"*.mp4"))
    else:
        file_list = glob(os.path.join(directory, f"*{media_ext}-[0-9]*"))
    logger.info("Found the following files:\n* %s", "\n* ".join(file_list))

    results = []
    for filename in tqdm(
        file_list,
        total=len(file_list),
        desc="Video Imports",
        dynamic_ncols=True,
        position=0,
        ascii=False,
    ):
        try:
            media_id = _upload_video(
                tator_api,
                filename,
                project_id,
                media_type,
                file_ids,
                section,
                vessel_name,
                skip_decrypt,
            )
        except Exception:
            logger.error("Failed to upload '%s'", os.path.basename(filename), exc_info=True)
        else:
            if media_id:
                results.append(media_id)
    return results



import argparse
import datetime
from glob import glob
import io
import logging
from math import floor, isnan
import os
from pprint import pformat
import requests
import subprocess
import tempfile
from typing import Any, Dict, List, Optional, Union

import tator
from tator.openapi.tator_openapi import TatorApi
from tator.openapi.tator_openapi.models import File
from tqdm import tqdm

from hms_import.util import LAND_DATE_PLACEHOLDER, SAIL_DATE_PLACEHOLDER, wait_for_thumbs


if __name__ != "__main__":
    logger = logging.getLogger(__name__)

GPS_EXT = ".log"
LOGS_TO_IGNORE = ["tracelog", "eventlog"]


def decrypt_file(data_dir: str, output_dir: str, private_key_path: Optional[str]) -> bool:
    retval = False
    if not private_key_path or not os.path.isfile(private_key_path):
        logger.error("Called decrypt_file with nonexistent private key path '%s'", private_key_path)
        return retval
    args = [
        "/scripts/FileProcessor",
        "--decrypt",
        private_key_path,
        "--verbose",
        output_dir,
        data_dir,
    ]
    try:
        with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as process:
            if process.stdout is not None:
                for line in io.TextIOWrapper(process.stdout, newline=""):
                    logger.info(line.strip())
    except Exception:
        logger.warning(f"Attempt to decrypt contents of '{data_dir}' unsuccessful", exc_info=True)
    else:
        retval = True

    return retval


def decrypt_media(
    data_dir: str, output_dir: str, toc_extension: str, private_key_path: Optional[str]
) -> Dict[str, Any]:
    if not decrypt_file(data_dir, output_dir, private_key_path):
        return {}

    toc_files = glob(os.path.join(output_dir, f"*{toc_extension}"))
    if len(toc_files) != 1:
        logger.warning("Found %d toc files, expect exactly one, skipping", len(toc_files))
        return {}

    toc_filename = toc_files[0]
    with open(toc_filename) as fp:
        parts = fp.read().strip().split(",")

    if len(parts) != 3:
        logger.warning("Toc file %s malformed, skipping", toc_files)
        return {"toc_filename": toc_filename}

    decrypted_filename, start_time_ms, end_time_ms = parts

    return {
        "toc_filename": toc_filename,
        "toc_start": datetime.datetime.fromtimestamp(
            int(start_time_ms) / 1000, tz=datetime.timezone.utc
        ),
        "toc_end": datetime.datetime.fromtimestamp(
            int(end_time_ms) / 1000, tz=datetime.timezone.utc
        ),
        "decrypted_name": decrypted_filename,
    }


def convert_int(value: str) -> int:
    try:
        return int(value)
    except Exception:
        pass
    return 0


def convert_float(value: str) -> float:
    try:
        converted_float = float(value)
    except Exception:
        converted_float = 0.0
    else:
        if isnan(converted_float):
            converted_float = 0.0
    return converted_float


def convert_datetime(value: str) -> datetime.datetime:
    try:
        return datetime.datetime.strptime(value, "%y%m%d%H%M%S%z")
    except Exception:
        pass
    return datetime.datetime.min


def _line_to_attrs(line: str) -> Union[Dict[str, Any], None]:
    try:
        (
            date_col,
            time_col,
            latitude_col,
            longitude_col,
            satellite_error_col,
            speed_col,
            heading_col,
            video_col,
            pressure1_col,
            set_rotation1_col,
            retrieval_rotation1_col,
            _,
        ) = line.split(",")
    except ValueError:
        logger.debug("Could not split line '%s' 12 times as desired, skipping", line, exc_info=True)
        return None
    return {
        "Satellite Count": convert_int(satellite_error_col),
        "Knots": convert_float(speed_col),
        "Heading": convert_float(heading_col),
        "Datecode": convert_datetime(f"{date_col}{time_col}Z"),
        "Position": [convert_float(longitude_col), convert_float(latitude_col)],
        "Video": convert_int(video_col),
        "Pressure 1": convert_float(pressure1_col),
        "Set Rotation 1": convert_float(set_rotation1_col),
        "Retrieval Rotation 1": convert_float(retrieval_rotation1_col),
    }


def decrypt_metadata(
    data_dir: str, output_dir: str, private_key_path: Optional[str]
) -> Optional[str]:
    # Decrypt the file
    gps_file_path = None
    if not decrypt_file(data_dir, output_dir, private_key_path):
        return gps_file_path

    # Make sure there is exactly one file with the desired extension
    gps_list = glob(os.path.join(output_dir, f"*{GPS_EXT}"))
    if len(gps_list) == 1:
        gps_file_path = gps_list[0]
    return gps_file_path


def parse_metadata(gps_file_path: str) -> List[Dict[str, Any]]:
    # Convert the lines of the file to attributes
    with open(gps_file_path) as fp:
        metadata = [_line_to_attrs(line) for line in fp.readlines()]

    return [point for point in metadata if point]


def main(
    *,
    tator_api: TatorApi,
    work_dir: str,
    project_id: int,
    media_ids: List[int],
    toc_extension: str,
    state_type: int,
    image_type: int,
    multi_type: int,
    private_key_path: Optional[str] = None,
) -> Optional[int]:
    media_list = tator_api.get_media_list_by_id(project_id, media_id_query={"ids": media_ids})
    section_name = ""
    section_id = -1
    media_type_id = None
    media_type = None
    summary_image_id = None
    if media_list:
        media = media_list[0]
        media_type_id = media.type
        media_type = tator_api.get_media_type(media_type_id)
        tator_user_sections = media.attributes["tator_user_sections"]
        section_list = tator_api.get_section_list(project_id)
        if isinstance(section_list, list):
            for sec in section_list:
                if sec.tator_user_sections == tator_user_sections:
                    section_name = sec.name
                    section_id = sec.id
                    break

    image_list = tator_api.get_media_list(project_id, section=section_id, type=image_type)
    if image_list:
        summary_image_id = image_list[0].id

    desired_streaming_resolutions = set()
    if hasattr(media_type, "streaming_config") and media_type.streaming_config:
        desired_streaming_resolutions.update(
            [config.resolution for config in media_type.streaming_config]
        )
    datetime_lookup = {}
    starts_and_ids = []
    wait_list = []

    # Decrypt and transcode all media files first
    logger.info("Importing media")
    sail_date = datetime.date.max
    land_date = datetime.date.min
    vessel_name = None
    for media in media_list:
        # Idempotency check to determine if a file needs to be downloaded and decrypted
        attributes = {}
        needs_transcode = True
        if (
            hasattr(media, "media_files")
            and media.media_files
            and hasattr(media.media_files, "streaming")
            and media.media_files.streaming
        ):
            needs_transcode = False
            for attr, converter in [
                ("toc_start", datetime.datetime.fromisoformat),
                ("toc_end", datetime.datetime.fromisoformat),
                ("decrypted_name", str),
            ]:
                if attr in getattr(media, "attributes", {}):
                    attributes[attr] = converter(media.attributes[attr])
                else:
                    needs_transcode = True

            if not needs_transcode:
                extant_resolutions = set(
                    [config.resolution[0] for config in media.media_files.streaming]
                )
                if extant_resolutions != desired_streaming_resolutions:
                    needs_transcode = True

        # Initialize with None and update with media object if `needs_transcode` is False or if
        # the decryption is successful, otherwise it will remain None and not be considered for
        # multiview generation
        updated_media = None
        if needs_transcode:
            with tempfile.TemporaryDirectory(dir=work_dir) as data_dir, tempfile.TemporaryDirectory(
                dir=work_dir
            ) as output_dir:
                # Get download info for encrypted file
                encrypted_path = media.attributes.get("encrypted_path", None)
                if encrypted_path:
                    download_info = tator_api.get_download_info(
                        project_id, {"keys": [encrypted_path]}
                    )
                    if not (
                        download_info
                        and isinstance(download_info, list)
                        and download_info[0].key == encrypted_path
                        and download_info[0].url
                    ):
                        raise RuntimeError(
                            f"Could not get download info for {encrypted_path=}; received:\n"
                            f"{pformat(download_info)}"
                        )

                    # Download the encrypted video file
                    encrypted_path = os.path.join(data_dir, media.name)
                    response = requests.get(download_info[0].url, stream=True)
                    with open(encrypted_path, "wb") as fp:
                        for chunk in response.iter_content(chunk_size=10485760):  # 10 MiB
                            fp.write(chunk)

                    # Decrypt and transcode the file
                    attributes = decrypt_media(
                        data_dir, output_dir, toc_extension, private_key_path
                    )
                    if not attributes:
                        logger.warning("Could not decrypt %s, skipping transcode", media.name)
                        continue

                    # Upload toc file, if appropriate
                    toc_filename = attributes.pop("toc_filename", None)
                    if toc_filename:
                        try:
                            for _, response in tator.util.upload_attachment(
                                tator_api, media.id, toc_filename
                            ):
                                pass
                        except Exception:
                            logger.warning(
                                "Could not upload toc file %s", toc_filename, exc_info=True
                            )
                        else:
                            logger.debug(
                                "Uploaded toc file %s as an attachment on media %d",
                                toc_filename,
                                media.id,
                            )
                    source_dir = output_dir
                    try:
                        tator_api.update_media(media.id, media_update={"attributes": attributes})
                    except Exception:
                        logger.warning(
                            "Could not update media %d with attributes %s", media.id, attributes
                        )
                else:
                    source_dir = work_dir
                    attributes = {
                        "toc_start": datetime.datetime.fromisoformat(media.attributes["toc_start"]),
                        "toc_end": datetime.datetime.fromisoformat(media.attributes["toc_end"]),
                        "decrypted_name": media.attributes["decrypted_name"],
                    }

                path = os.path.join(source_dir, attributes["decrypted_name"])
                if os.path.isfile(path):
                    filename = attributes["decrypted_name"]
                    pbar = tqdm(
                        desc=filename, dynamic_ncols=True, position=1, total=100, ascii=False
                    )
                    last_progress = 0
                    logger.info("Uploading media for %s", filename)
                    for p, _ in tator.util.upload_media(
                        api=tator_api,
                        type_id=media.type,
                        path=path,
                        md5=tator.util.md5sum(path),
                        fname=filename,
                        media_id=media.id,
                        timeout=120,
                    ):
                        increment = max(p - last_progress, 1)
                        pbar.update(increment)
                        last_progress = p
                    wait_list.append(media.id)
                    updated_media = tator_api.get_media(media.id)
                else:
                    logger.error("Error decrypting %s", attributes["decrypted_name"])
                    err_log_path = f"{path}.log"

                    if os.path.isfile(err_log_path):
                        try:
                            for _, response in tator.util.upload_attachment(
                                tator_api, media.id, err_log_path
                            ):
                                pass
                        except Exception:
                            logger.warning(
                                "Could not upload error log file %s", err_log_path, exc_info=True
                            )
                        else:
                            logger.debug(
                                "Uploaded error log file %s as an attachment on media %d",
                                err_log_path,
                                media.id,
                            )
                        with open(err_log_path) as fp:
                            for line in fp.readlines():
                                logger.error(line.strip())
                    else:
                        logger.error("Could not open log file at '%s'", err_log_path)
        else:
            updated_media = media

        # Store info for creating multiviews
        if updated_media:
            starts_and_ids.append((attributes["toc_start"], media.id))
            datetime_lookup[media.id] = {"obj": updated_media, **attributes}
            sail_date = min(sail_date, attributes["toc_start"].date())
            land_date = max(land_date, attributes["toc_end"].date())
            if not vessel_name:
                vessel_name = media.attributes.get("Vessel Name", None)

    starts_and_ids.sort()
    current_start, media_id = starts_and_ids[0]
    multi_lookup = {current_start: [media_id]}
    for start_dt, media_id in starts_and_ids[1:]:
        if (start_dt - current_start).total_seconds() > 5:
            multi_lookup[start_dt] = []
            current_start = start_dt
        multi_lookup[current_start].append(media_id)
    # Determine extant multiviews by comparing against id lists and remove from creation list
    media_id_sets = [(start, set(media_ids)) for start, media_ids in multi_lookup.items()]
    paginator = tator.util.get_paginator(tator_api, "get_media_list")
    kwargs = {"project": project_id, "type": multi_type}
    if section_id:
        kwargs["section"] = section_id
    page_iter = paginator.paginate(**kwargs)
    try:
        for page in page_iter:
            for multi in page:
                multi_id_set = set()
                if (
                    hasattr(multi, "media_files")
                    and hasattr(multi.media_files, "ids")
                    and multi.media_files.ids
                ):
                    multi_id_set.update(multi.media_files.ids)
                if multi_id_set:
                    for start, media_id_set in media_id_sets:
                        if multi_id_set == media_id_set:
                            multi_lookup.pop(start, None)
                            break
    except RuntimeError:
        # Paginator has a bug in handling zero results
        pass

    # Create non-existent multiviews
    if wait_for_thumbs(tator_api, wait_list):
        logger.info("Creating %d multiviews", len(multi_lookup))
        for start, media_ids in multi_lookup.items():
            # Get latest start time
            multi_start = max(datetime_lookup[media_id]["toc_start"] for media_id in media_ids)

            # Sort media by filename for consistent display order
            media_objects = [datetime_lookup[media_id]["obj"] for media_id in media_ids]
            media_objects.sort(key=lambda ele: ele.name)
            ordered_media_ids = [m.id for m in media_objects]

            # Calculate frame offsets
            frame_offsets = [
                floor(
                    (multi_start - datetime_lookup[media.id]["toc_start"]).total_seconds()
                    * media.fps
                )
                if media.fps
                else 0
                for media in media_objects
            ]

            try:
                tator.util.make_multi_stream(
                    api=tator_api,
                    type_id=multi_type,
                    layout=[1, len(ordered_media_ids)],
                    name=start.isoformat(),
                    media_ids=ordered_media_ids,
                    section=section_name,
                    frame_offset=frame_offsets,
                )
            except Exception:
                logger.warning(
                    "Failed to make multi for start time %s", start.isoformat(), exc_info=True
                )

    # Update section and media with sail and land dates
    new_section_name = section_name
    new_section_name = new_section_name.replace(
        SAIL_DATE_PLACEHOLDER, sail_date.strftime("%Y-%m-%d")
    )
    new_section_name = new_section_name.replace(
        LAND_DATE_PLACEHOLDER, land_date.strftime("%Y-%m-%d")
    )
    try:
        tator_api.update_section(section_id, section_update={"name": new_section_name})
    except Exception:
        logger.warning(
            "Could not change section name from %s to %s",
            section_name,
            new_section_name,
            exc_info=True,
        )
    bulk_update = {
        "attributes": {
            "Sail Date": sail_date,
            "Land Date": land_date,
            "Vessel Name": vessel_name,
        }
    }
    for mt in [media_type_id, multi_type, image_type]:
        try:
            tator_api.update_media_list(project_id, bulk_update, type=mt, section=section_id)
        except Exception:
            logger.warning(
                "Could not apply bulk update '%s' to MediaType %d", bulk_update, mt, exc_info=True
            )

    related_file_ids = set(
        int(file_id)
        for entry in datetime_lookup.values()
        for file_id in entry["obj"].attributes.get("related_files", "").split(",")
        if file_id.isnumeric()  # Protects against an empty string
    )

    # For idempotency, delete any existing state specs on the media, one media at a time to avoid
    # too many deletions
    for media_id in datetime_lookup.keys():
        try:
            response = tator_api.delete_state_list(project_id, media_id=[media_id])
        except Exception:
            logger.warning("Could not delete existing states on media %d", media_id, exc_info=True)

    metadata = []
    if related_file_ids:
        for file_id in related_file_ids:
            file_obj = tator_api.get_file(file_id)
            if not isinstance(file_obj, File):
                logger.warning("Could not find file with id %d, skipping", file_id)
                continue
            if not isinstance(file_obj.name, str):
                logger.warning("Could not get name of file %d, skipping", file_id)
                continue
            download_info = tator_api.get_download_info(
                file_obj.project, download_info_spec={"keys": [file_obj.path]}
            )

            if not (download_info and isinstance(download_info, list) and download_info[0].url):
                logger.warning(
                    "Could not get download info for '%s' (%d), skipping", file_obj.name, file_id
                )
                continue
            else:
                download_url = download_info[0].url

            # Make working directories
            with tempfile.TemporaryDirectory(dir=work_dir) as data_dir, tempfile.TemporaryDirectory(
                dir=work_dir
            ) as output_dir:
                # Download the encrypted GPS file
                encrypted_path = os.path.join(data_dir, file_obj.name)
                response = requests.get(download_url, stream=True)
                with open(encrypted_path, "wb") as fp:
                    for chunk in response.iter_content(chunk_size=128):
                        fp.write(chunk)

                log_filename = decrypt_metadata(data_dir, output_dir, private_key_path)
                if log_filename:
                    if any(kw in file_obj.name for kw in LOGS_TO_IGNORE):
                        logger.debug(
                            "Not parsing log file '%s'; it does not contain GPS data", file_obj.name
                        )
                    else:
                        metadata.extend(parse_metadata(log_filename))
                    try:
                        for _ in tator.util.upload_attachment(
                            api=tator_api, media=summary_image_id, path=log_filename
                        ):
                            pass
                    except Exception:
                        pass
    else:
        gps_list = glob(os.path.join(work_dir, f"*primary_sensor*{GPS_EXT}"))
        # Convert the lines of the file to attributes
        for gps_file in gps_list:
            with open(gps_file) as fp:
                file_points = [_line_to_attrs(line) for line in fp.readlines()]
            metadata.extend(point for point in file_points if point)

    state_specs = []
    # Create a state spec for each point/media pair, where the point timestamp occurs between
    # the start and stop timestamps of the media object
    for point in metadata:
        for media_id, media_info in datetime_lookup.items():
            if (
                media_info["toc_start"] <= point["Datecode"] <= media_info["toc_end"]
                and media_info["obj"].fps
            ):
                delta = (point["Datecode"] - media_info["toc_start"]).total_seconds()
                frame = floor(delta * media_info["obj"].fps)
                state_specs.append(
                    {
                        "type": state_type,
                        "media_ids": [media_id],
                        "frame": frame,
                        "attributes": point,
                    }
                )

    for response in tator.util.chunked_create(
        tator_api.create_state_list, project_id, body=state_specs
    ):
        logger.info(response.message)

    return summary_image_id


if __name__ == "__main__":
    from hms_import.util import HmsLogHandler

    parser = argparse.ArgumentParser(
        description="Script for decrypting and transcoding encrypted video files"
    )
    # Get updated media file metadata
    tator.get_parser(parser)
    parser.add_argument("--work-dir", type=str, required=True)
    parser.add_argument("--project-id", type=int, required=True)
    parser.add_argument("--media-ids", nargs="+", type=int, required=True)
    parser.add_argument("--toc-extension", type=str, required=True)
    parser.add_argument("--state-type", type=int, required=True)
    parser.add_argument("--image-type", type=int, required=True)
    parser.add_argument("--multi-type", type=int, required=True)
    parser.add_argument("--private-key-path", type=str, required=False)
    args = parser.parse_args()
    argdict = vars(args)
    host = argdict.pop("host")
    token = argdict.pop("token")
    api = tator.get_api(host=host, token=token)
    argdict["tator_api"] = api

    with HmsLogHandler(
        tator_api=api,
        console_log_level=logging.INFO,
        log_filename=os.path.splitext(os.path.basename(__file__))[0],
    ) as log_handler:
        logger = logging.getLogger(__name__)
        log_handler.summary_image_id = main(**argdict)



import tator

from hms_import.b3.upload import run_b3_upload
from hms_import.util import HmsLogHandler


def main(*, host: str, token: str, console_log_level: int, **kwargs):
    tator_api = tator.get_api(host=host, token=token)
    with HmsLogHandler(tator_api=tator_api, console_log_level=console_log_level) as log_handler:
        log_handler.summary_image_id = run_b3_upload(tator_api=tator_api, **kwargs)



import datetime
from glob import iglob
import logging
import os
from typing import Any, Dict, List, Optional, Tuple

import tator
from tator.openapi.tator_openapi import (
    CreateListResponse,
    File,
    FileType,
    MediaType,
    Section,
    TatorApi,
)
from tqdm import tqdm

from hms_import.b3.generate_states import generate_multis_and_states
from hms_import.util import build_section_name, safe_get_type, section_from_name, wait_for_thumbs
from hms_import.summary_image import create_summary_image

logger = logging.getLogger(__name__)
DATETIME_FORMAT = "%Y%m%dT%H%M%SZ"


def upload_and_import_videos(
    tator_api: TatorApi,
    video_list: List[str],
    media_type: MediaType,
    summary_type_id: int,
    vessel_name: str,
    file_list: List[File],
    section_name: str,
) -> Tuple[List[int], Dict[str, Any], str]:
    """Returns list of created media ids"""
    created_ids = []
    media_type_id = media_type.id
    project_id = media_type.project
    shared_attributes = {
        "Vessel Name": vessel_name,
        "Sail Date": datetime.date.max,
        "Land Date": datetime.date.min,
        "related_files": ",".join(str(file_obj.id) for file_obj in file_list),
    }

    # Collect media_specs
    media_specs = []
    for video_path in video_list:
        try:
            filename = os.path.basename(video_path)
            filename_parts = os.path.splitext(filename)[0].split("-")
            start_datetime = datetime.datetime.min
            if len(filename_parts) == 4:
                start_date, start_time = filename_parts[2:4]
                try:
                    start_datetime = datetime.datetime.strptime(
                        f"{start_date}T{start_time}", DATETIME_FORMAT
                    ).replace(tzinfo=datetime.timezone.utc)
                except Exception:
                    logger.warning("Could not parse datetime from filename '%s'", filename)
            if start_datetime != datetime.datetime.min:
                shared_attributes["Sail Date"] = min(
                    shared_attributes["Sail Date"], start_datetime.date()
                )
            if start_datetime != datetime.datetime.min:
                shared_attributes["Land Date"] = max(
                    shared_attributes["Land Date"], start_datetime.date()
                )
            vid_md5 = tator.util.md5sum(video_path)

            # Store the media spec and video path for creation
            media_specs.append(
                (
                    video_path,
                    {
                        "attributes": {"toc_start": start_datetime},
                        "name": filename,
                        "type": media_type_id,
                        "section": section_name,
                        "md5": vid_md5,
                    },
                )
            )
        except Exception:
            logger.warning("Encountered exception while processing '%s', skipping", video_path)

    # Rename section
    logger.info("renaming section")
    section = section_from_name(tator_api, project_id, section_name)
    new_section_name = section_name
    section_id = -1
    if isinstance(section, Section):
        section_id = section.id
        new_section_name = build_section_name(
            vessel_name=shared_attributes["Vessel Name"],
            sail_date=shared_attributes["Sail Date"],
            land_date=shared_attributes["Land Date"],
        )
        try:
            response = tator_api.update_section(
                section.id, section_update={"name": new_section_name}
            )
        except Exception:
            new_section_name = section_name

    # Upload files and create media objects
    logger.info("uploading files")
    for video_path, media_spec in tqdm(
        media_specs, desc="Video Imports", dynamic_ncols=True, position=0, ascii=False
    ):
        try:
            media_spec["section"] = new_section_name
            media_spec["attributes"].update(shared_attributes)
            response = tator_api.create_media_list(project_id, body=[media_spec])
            if isinstance(response, CreateListResponse) and response.id:
                filename = media_spec["name"]
                vid_md5 = media_spec["md5"]
                created_ids.append(response.id[0])
                pbar = tqdm(desc=filename, dynamic_ncols=True, position=1, total=100, ascii=False)
                last_progress = 0
                for p, _ in tator.util.upload_media(
                    api=tator_api,
                    type_id=media_type_id,
                    path=video_path,
                    md5=vid_md5,
                    section=new_section_name,
                    fname=filename,
                    media_id=created_ids[-1],
                    timeout=120,
                ):
                    increment = max(p - last_progress, 1)
                    pbar.update(increment)
                    last_progress = p

            else:
                logger.warning("Could not create media object for '%s', skipping", video_path)
        except Exception:
            logger.warning("Encountered exception while processing '%s', skipping", video_path)

    # Update summary image, if any, with shared attributes
    shared_attributes.pop("related_files", None)
    try:
        tator_api.update_media_list(
            project_id,
            media_bulk_update={"attributes": shared_attributes},
            type=summary_type_id,
            section=section_id,
        )
    except Exception:
        logger.warning("Failed to update summary image with attributes %s", shared_attributes)

    return created_ids, shared_attributes, new_section_name


def upload_sensor_data(
    tator_api: TatorApi,
    sensor_list: List[str],
    file_type: FileType,
) -> List[File]:
    file_type_id = file_type.id
    file_list = []
    for sensor_path in tqdm(
        sensor_list, dynamic_ncols=True, desc="Sensor Import", position=0, ascii=False
    ):
        filename = os.path.basename(sensor_path)
        try:
            pbar = tqdm(desc=filename, dynamic_ncols=True, position=1, total=100, ascii=False)
            last_progress = 0
            response = None
            for p, response in tator.util.upload_generic_file(
                api=tator_api,
                file_type=file_type_id,
                path=sensor_path,
                description="Raw sensor data",
                name=filename,
                timeout=120,
            ):
                increment = max(p - last_progress, 1)
                pbar.update(increment)
                last_progress = p
        except Exception:
            logger.warning("Encountered exception while processing '%s', skipping", sensor_path)
            continue
        if isinstance(response, File):
            file_list.append(response)
    return file_list


def run_b3_upload(
    *,
    tator_api: TatorApi,
    media_type_id: int,
    file_type_id: int,
    multi_type_id: int,
    state_type_id: int,
    image_type_id: int,
    directory: str,
    hdd_sn: Optional[str] = None,
) -> Optional[int]:
    """
    :tator_api: The TatorApi object to use for interactions with Tator
    :media_type_id: The unique ID of the type of video to create
    :file_type_id: The unique ID of the type of file to create for storing GPS data
    :multi_type_id: The unique ID of the type of multiview to create
    :state_type_id: The unique ID of the type of State to create
    :image_type_id: The unique ID of the type of summary image to create
    :directory: The folder containing the files to import
    :hdd_sn: The hard drive serial number
    :returns: The summary image id
    """
    summary_id = None

    # Validate the given media and file types, abort if they do not exist or are incompatible
    media_type = safe_get_type(media_type_id, tator_api.get_media_type)
    file_type = safe_get_type(file_type_id, tator_api.get_file_type)
    if media_type is None:
        logger.error("Could not get media type %d from Tator, aborting", media_type_id)
        return summary_id
    if file_type is None:
        logger.error("Could not get file type %d from Tator, aborting", file_type_id)
        return summary_id
    if media_type.project != file_type.project:
        logger.error(
            "Received MediaType %d and FileType %d, which are from different projects, aborting",
            media_type_id,
            file_type_id,
        )
        return summary_id

    # Locate media for import and create summary image
    video_list = list(iglob(os.path.join(directory, f"*.mp4")))
    summary_id, section_name, vessel_name = create_summary_image(
        tator_api=tator_api,
        media_type=image_type_id,
        import_type="B3",
        directory=directory,
        hdd_sn=hdd_sn,
    )

    # Locate sensor data for import
    sensor_list = list(iglob(os.path.join(directory, f"*.log")))
    if sensor_list:
        file_list = upload_sensor_data(tator_api, sensor_list, file_type)
    else:
        file_list = []
        logger.warning("No sensor data found, only videos will be imported")

    if video_list:
        logger.debug("importing videos")
        created_ids, shared_attrs, new_section_name = upload_and_import_videos(
            tator_api, video_list, media_type, image_type_id, vessel_name, file_list, section_name
        )
    else:
        logger.error("No media found, aborting")
        return summary_id

    # Generate associated multiviews and GPS States
    if wait_for_thumbs(tator_api, created_ids):
        generate_multis_and_states(
            tator_api=tator_api,
            media_type_id=media_type_id,
            multi_type_id=multi_type_id,
            state_type_id=state_type_id,
            section_name=new_section_name,
            attrs=shared_attrs,
        )
    return summary_id



from argparse import ArgumentParser, RawDescriptionHelpFormatter
import logging

import tator


def get_parser():
    parser = ArgumentParser(
        description="Script for importing video and metadata in O2 and B3 formats.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        help="Changes the console log level from INFO to WARNING; defers to --verbose",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Changes the console log level from INFO to DEBUG; takes precedence over --quiet",
    )
    cmd_parser = parser.add_subparsers(title="Commands", dest="command")

    b3_parser = cmd_parser.add_parser(
        "b3-upload", help="Imports video and GPS files from unlocked LUKS-encrypted device"
    )
    tator.get_parser(b3_parser)
    b3_parser.add_argument("--media-type-id", type=int)
    b3_parser.add_argument("--file-type-id", type=int)
    b3_parser.add_argument("--multi-type-id", type=int)
    b3_parser.add_argument("--state-type-id", type=int)
    b3_parser.add_argument("--image-type-id", type=int)
    b3_parser.add_argument("--directory", type=str)
    b3_parser.add_argument("--hdd-sn", type=str, required=False)

    o2_parser = cmd_parser.add_parser(
        "o2-upload", help="Script for uploading raw, encrypted video files"
    )
    o2_parser.add_argument(
        "config_file", type=str, help=f"The configuration .ini file used to initialize {__name__}."
    )

    return parser


def main() -> None:
    # Parse arguments
    parser = get_parser()
    args = parser.parse_args()
    argdict = vars(args)

    # Always log everything to file, set console log level based on `--quiet` and `--verbose` flags
    argdict["console_log_level"] = logging.INFO
    if argdict.pop("quiet"):
        argdict["console_log_level"] = logging.WARNING
    if argdict.pop("verbose"):
        argdict["console_log_level"] = logging.DEBUG

    # Import the desired main function
    command = argdict.pop("command")
    if command == "o2-upload":
        from hms_import.o2 import main
    elif command == "b3-upload":
        from hms_import.b3 import main
    else:
        raise RuntimeError(f"Received unhandled command '{command}'")

    main(**argdict)



from collections import defaultdict
import datetime
import logging
from math import floor
import os
import requests
import tempfile
from typing import Any, Dict

from natsort import natsorted
import pynmea2
import tator
from tator.openapi.tator_openapi import File, MediaType, TatorApi
from tqdm import tqdm

from hms_import.util import section_from_name

DATETIME_FORMAT = "%Y%m%dT%H%M%SZ"
logger = logging.getLogger(__file__)


def _line_to_attrs(line):
    stripped_line = line.strip("\x00")
    if stripped_line:
        try:
            message = pynmea2.parse(f"$GPRMC,{stripped_line}")
            return {
                "Knots": message.spd_over_grnd,
                "Heading": message.true_course,
                "Datecode": message.datetime,
                "Position": [message.lon, message.lat],
            }
        except Exception:
            logger.debug("Skipping malformed gps entry: '%s'", line)
    return None


def points_from_file(gps_data_path):
    with open(gps_data_path, "r") as fp:
        lines = fp.readlines()
    metadata = [_line_to_attrs(line) for line in lines]
    return [point for point in metadata if point is not None]


def generate_multis_and_states(
    tator_api: TatorApi,
    media_type_id: int,
    multi_type_id: int,
    state_type_id: int,
    section_name: str,
    attrs: Dict[str, Any],
):
    try:
        multi_type = tator_api.get_media_type(multi_type_id)
    except Exception:
        logger.error(
            "Could not find MediaType '%d' in Tator, aborting!", multi_type_id, exc_info=True
        )
        return
    if not isinstance(multi_type, MediaType):
        return
    if multi_type.project:
        project_id = multi_type.project
    else:
        return

    section = section_from_name(tator_api, project_id, section_name)
    if section is None:
        logger.error(
            "Found no section named '%s' in project '%d', aborting!", section_name, project_id
        )
        return

    media_list = tator_api.get_media_list(
        project=project_id, section=section.id, type=media_type_id
    )
    if not (isinstance(media_list, list) or media_list):
        logger.error(
            "Found no media in section '%s' in project '%d', aborting!", section_name, project_id
        )
        return

    project_id = media_list[0].project
    datetime_lookup = {}
    multi_lookup = defaultdict(list)
    state_file_ids = set()

    for media in media_list:
        try:
            # Get info to create states
            state_file_ids.update(
                int(fid)
                for fid in media.attributes.get("related_files", "").split(",")
                if fid.isnumeric()
            )

            toc_start = datetime.datetime.fromisoformat(media.attributes["toc_start"])
            duration = datetime.timedelta(seconds=media.num_frames / media.fps)
            datetime_lookup[media.id] = {
                "fps": media.fps,
                "toc_start": toc_start,
                "toc_end": toc_start + duration,
            }

            # Get info to create multiviews
            if not media.name:
                logger.warning("Could not get filename for media '%d', skipping", media.id)
                continue
            filename = media.name
            multi_lookup[toc_start].append((filename, media.id))
        except Exception:
            logger.error(
                "Encountered exception while processing '%d', skipping", media.id, exc_info=True
            )

    # Make multiviews from pairs of concurrent videos
    for start_dt, media_lst in tqdm(
        list(multi_lookup.items()),
        desc="Creating Multis",
        dynamic_ncols=True,
        position=0,
        ascii=False,
    ):
        sorted_ids = [ele[-1] for ele in natsorted(media_lst)]
        tator.util.make_multi_stream(
            api=tator_api,
            type_id=multi_type_id,
            layout=[1, 2],
            name=start_dt.strftime(DATETIME_FORMAT),
            media_ids=sorted_ids,
            section=section.name,
        )

    try:
        tator_api.update_media_list(
            project=project_id,
            media_bulk_update={"attributes": attrs},
            type=multi_type_id,
            section=section.id,
        )
    except Exception:
        logger.warning("Failed to update attributes on multiviews, moving on")

    # Generate states
    state_specs = []
    for file_id in state_file_ids:
        try:
            file_obj = tator_api.get_file(file_id)
        except Exception:
            logger.error("Could not get File '%d', skipping", file_id, exc_info=True)
            continue
        if not isinstance(file_obj, File):
            logger.warning("Could not find file with id %d, skipping", file_id)
            continue
        if not isinstance(file_obj.name, str):
            logger.warning("Could not get name of file %d, skipping", file_id)
            continue
        download_info = tator_api.get_download_info(
            file_obj.project, download_info_spec={"keys": [file_obj.path]}
        )

        if not (download_info and isinstance(download_info, list) and download_info[0].url):
            logger.warning(
                "Could not get download info for '%s' (%d), skipping", file_obj.name, file_id
            )
            continue
        else:
            download_url = download_info[0].url

        # Make working directories
        with tempfile.TemporaryDirectory() as data_dir:
            # Download the encrypted GPS file
            gps_data_path = os.path.join(data_dir, file_obj.name)
            response = requests.get(download_url, stream=True)
            with open(gps_data_path, "wb") as fp:
                for chunk in response.iter_content(chunk_size=128):
                    fp.write(chunk)

            metadata = points_from_file(gps_data_path)

        # Create a state spec for each point/media pair, where the point timestamp occurs between
        # the start and stop timestamps of the media object
        for point in metadata:
            for media_id, media_info in datetime_lookup.items():
                if (
                    media_info["toc_start"] <= point["Datecode"] <= media_info["toc_end"]
                    and media_info["fps"]
                ):
                    delta = (point["Datecode"] - media_info["toc_start"]).total_seconds()
                    frame = floor(delta * media_info["fps"])
                    state_specs.append(
                        {
                            "type": state_type_id,
                            "media_ids": [media_id],
                            "frame": frame,
                            "attributes": point,
                        }
                    )

    for response in tator.util.chunked_create(
        tator_api.create_state_list, project_id, body=state_specs
    ):
        logger.info(response.message)
